"""
May 25, 2018

Marine Cadastre Project
www.marinecadastre.gov

NOAA Office for Coastal Management
1234 South Hobson Avenue
Charleston, SC 29405
843-740-1200
ocm.mmc@noaa.gov

Software design and development by:
RPS
55 Village Square Drive
Wakefield, RI 02879
"""

import arcpy, os, sys, traceback, datetime, math, re


class Toolbox(object):
    def __init__(self):
        self.label = "AIS Utilities 2018 Pro"
        self.alias = "AISUtilities2018Pro"

        self.tools = [AIS_TrackBuilder, AIS_CSVtoFGDB]


class AIS_TrackBuilder(object):
    def __init__(self):
        self.label = "AIS TrackBuilder"
        self.description = "Takes point features containing a date/time value and converts them to a Polyline feature class in sequential order. The track line can be segmented by either a user defined maximum distance or time between points, or by a user defined time interval."
        self.canRunInBackground = False

    def getParameterInfo(self):
        param0 = arcpy.Parameter(
            displayName="Input Points",
            name="Inputs_Points",
            datatype="DEFeatureClass",
            parameterType="Required",
            direction="Input")        
        param0.filter.list = ["Point"]
        
        param1 = arcpy.Parameter(
            displayName="ID Field",
            name="ID_Field",
            datatype="Field",
            parameterType="Required",
            direction="Input")  
        param1.filter.list = ['Short', 'Long']        
        param1.parameterDependencies = [param0.name]      
        #param1.value = "MMSI"
        
        param2 = arcpy.Parameter(
            displayName=r"Date/Time Field",
            name="DateTimeField",
            datatype="Field",
            parameterType="Required",
            direction="Input") 
        param2.filter.list = ['Date']       
        param2.parameterDependencies = [param0.name]
        #param2.value = "BaseDateTime"
        
        param3 = arcpy.Parameter(
            displayName="Output Workspace",
            name="Output_Workspace",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input") 
        param3.filter.list = ['Local Database']
        
        param4 = arcpy.Parameter(
            displayName="Output Name",
            name="OutputName",
            datatype="GPString",
            parameterType="Required",
            direction="Input") 
        
        param5 = arcpy.Parameter(
            displayName="Break Track Lines Using",
            name="TrackBreakOption",
            datatype="GPString",
            parameterType="Required",
            direction="Input")
        param5.filter.type = "ValueList"
        param5.filter.list = ["Maximum Time and Distance","Time Interval"]
        param5.value = "Maximum Time and Distance"
        
        param6 = arcpy.Parameter(
            displayName="Maximum Time (minutes)",
            name="MaxTime",
            datatype="GPLong",
            parameterType="Optional",
            direction="Input") 
        
        param7 = arcpy.Parameter(
            displayName="Maximum Distance (miles)",
            name="MaxDist",
            datatype="GPLong",
            parameterType="Optional",
            direction="Input") 
        
        param8 = arcpy.Parameter(
            displayName="Time Interval (hours)",
            name="TimeInt",
            datatype="GPLong",
            parameterType="Optional",
            direction="Input") 
 
        param9 = arcpy.Parameter(
            displayName="Fields to include as track line attributes",
            name="AddFieldList",
            datatype="Field",
            parameterType="Optional",
            direction="Input",
            multiValue=True)
        param9.filter.list = ['Short', 'Long', 'Single', 'Double', 'Text']       
        param9.parameterDependencies = [param0.name]
                
        params = [param0,param1,param2,param3,param4,param5,param6,param7,param8,param9]
        return params

    def isLicensed(self):
        dicInstallInfo = arcpy.GetInstallInfo()
        fArcVersion = float(dicInstallInfo["Version"].split(".")[0] + "." + dicInstallInfo["Version"].split(".")[1])
        if fArcVersion > 1.0:    
            return True
        else:
            return False

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        
        lFieldOptions = []
        if parameters[0].value and not parameters[0].hasBeenValidated:
            if arcpy.Exists(parameters[0].value):
                desc = arcpy.Describe(parameters[0].value)
                fieldList = desc.fields
                for field in fieldList:
                    if field.name == "MMSI":
                        parameters[1].value = "MMSI"
        
        if not parameters[5].hasBeenValidated:
            if parameters[5].value == "Maximum Time and Distance":
                parameters[6].enabled = 1
                parameters[6].value= 30
                parameters[7].enabled = 1
                parameters[7].value= 1
                parameters[8].enabled = 0
                parameters[8].value= 0
            elif parameters[5].value == "Time Interval":
                parameters[6].enabled = 0
                parameters[6].value= 0
                parameters[7].enabled = 0
                parameters[7].value= 0
                parameters[8].enabled = 1
                parameters[8].value= 24
            else:
                parameters[6].enabled = 0
                parameters[6].value= 0
                parameters[7].enabled = 0
                parameters[7].value= 0
                parameters[8].enabled = 0
                parameters[8].value= 0            
        return
    
    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        
        #Validate output FC name if the workspace and output name are set
        if parameters[4].value != "" and parameters[3].value:# and not self.params[4].hasBeenValidated:
            bValidName = True
            
            if str(parameters[4].value) != str(arcpy.ValidateTableName(parameters[4].value,parameters[3].value)):
                bValidName = False
        
            if bValidName == False:
                parameters[4].setErrorMessage("Output name is not a valid feature class name")
    
        #Validate Split options
        if parameters[5].value == "Maximum Time and Distance":
            if parameters[6].value <= 0:
                parameters[6].setErrorMessage("Maximum Time must be greater than 0 minutes")
            if parameters[7].value <= 0:
                parameters[7].setErrorMessage("Maximum Distance must be greater than 0 miles")
    
        elif parameters[5].value == "Time Interval":
            if parameters[8].value <= 0:
                parameters[8].setErrorMessage("Time Interval must be greater than 0 hours")           
        return

    def execute(self, parameters, messages):
        arcpy.env.overwriteOutput = True

        try:
            self.CreateTracks(parameters)
             
        except arcpy.ExecuteError:
            for msg in range(0, arcpy.GetMessageCount()):
                if arcpy.GetSeverity(msg) == 2:
                    arcpy.AddReturnMessage(msg)
                    print(arcpy.GetMessage(msg))
         
        except:
            tb = sys.exc_info()[2]
            for i in range(0,len(traceback.format_tb(tb))):
                tbinfo = traceback.format_tb(tb)[i]
                msg = "   ERROR: \n   " + tbinfo + "   Error Info:\n   " + str(sys.exc_info()[1])

            arcpy.AddError(msg)
            print(msg)
         
        return
    




    ################################################
    ################################################
    ################################################
    def printit(self, inMessage):
        """Simple method to send messages to ESRI geoprocessing tool dialog, or print to screen if run from another method."""
        print(inMessage)
        arcpy.AddMessage(inMessage)
    
    def FormatCounts(self, iInCount):
        """Formats various counts to strings with comma separators, for use in dialog messages."""
        sOutCount = re.sub("(\d)(?=(\d{3})+(?!\d))", r"\1,", "%d" % iInCount)
        return sOutCount
    
    def CreateDataDict(self):
        """Process to read through the input point feature class.  Point information is stored in a Python dictionary in memory.
        Points are read using the Data Access Module search cursor."""
        self.printit("\n  Reading input point data...")
        iTotPoints = int(arcpy.GetCount_management(sInputFile).getOutput(0))
        arcpy.SetProgressor("step","Reading input point data...",0,iTotPoints,1)
    
        dataDicAIS = {}
        dataDicAttrs = {}
        
        if sID_Field == "MMSI":
            where = sID_Field+" > 0"
        else:
            where = None
            
        fields = ["SHAPE@XY",sID_Field,sDT_Field]
        for f in lFields:
            if f.name not in fields:
                fields.append(f.name)
         
        with arcpy.da.SearchCursor(sInputFile,fields,where,spatialRef_CURSOR,False,sql_clause=(None,'ORDER BY '+sID_Field+', '+sDT_Field+' ASC')) as rows:
            for row in rows:
                pnt = row[0]
    
                sID = row[1]
                dDateTime = row[2]                 
    
                if sID in dataDicAIS:
                    dataDicAIS[sID][dDateTime] = [pnt[0],pnt[1]]
                else:
                    dataDicAIS[sID] = {dDateTime:[pnt[0],pnt[1]]}
                 
                #store the FIRST points attributes for each of the additional fields included   
                if sID not in dataDicAttrs:
                    dataDicAttrs[sID] = {}
                    for iFieldIndex in range(3,len(fields)):
                        dataDicAttrs[sID][fields[iFieldIndex]] = row[iFieldIndex]
    
                arcpy.SetProgressorPosition()
         
        self.printit("    Read "+self.FormatCounts(iTotPoints)+" points")
        arcpy.ResetProgressor()     
        return dataDicAIS, dataDicAttrs
    
    def CreateOuputDataset(self):
        """Creates an empty feature class to store the track polylines.  Adds the required fields, based on the options selected by the user."""
        self.printit("\n  Building output track line feature class...")
        arcpy.SetProgressor("default","Building output track line feature class...")
         
        tmp = os.path.join('in_memory', 'template')
         
        arcpy.CreateFeatureclass_management('in_memory','template',"POLYLINE","","DISABLED","DISABLED",sInputFile)
        
        lOutFieldList = ["SHAPE@",sID_Field,"TrackStartTime","TrackEndTime"]        
        arcpy.AddField_management(tmp,sID_Field,"LONG",)
        arcpy.AddField_management(tmp,"TrackStartTime","DATE",)
        arcpy.AddField_management(tmp,"TrackEndTime","DATE",)
         
        #Add other fields
        for f in lFields:
            if f.name not in [sID_Field,sDT_Field]:
                arcpy.AddField_management(tmp,f.name, f.type, f.precision, f.scale, f.length, f.aliasName)
                lOutFieldList.append(f.name)
    
        arcpy.CreateFeatureclass_management(sOutWorkspace,sOutName,"POLYLINE",tmp,"DISABLED","DISABLED",sInputFile)
         
        arcpy.Delete_management(tmp)
        del tmp
         
        arcpy.ResetProgressor()
        
        return lOutFieldList
        
    
    def GetDistance(self, prevX,prevY,currX,currY):
        """Calculates the distance between two latitude/longitude coordinate pairs, using the Vincenty formula for ellipsoid based distances.
        Returns the distance in miles."""
        #Vincenty Formula
        #Copyright 2002-2012 Chris Veness
        #http://www.movable-type.co.uk/scripts/latlong-vincenty.html
         
        a = 6378137
        b = 6356752.314245
        f = 1/298.257223563
    
        L = math.radians(currX - prevX)
        U1 = math.atan((1-f)*math.tan(math.radians(prevY)))
        U2 = math.atan((1-f)*math.tan(math.radians(currY)))
        sinU1 = math.sin(U1)
        sinU2 = math.sin(U2)
        cosU1 = math.cos(U1)
        cosU2 = math.cos(U2)
    
        lam = L
        lamP = 9999999999
        iter_count = 0
        while abs(lam-lamP) > 1e-12 and iter_count <= 100:
            sinLam = math.sin(lam)
            cosLam = math.cos(lam)
            sinSigma = math.sqrt((cosU2*sinLam)*(cosU2*sinLam) + (cosU1*sinU2-sinU1*cosU2*cosLam)*(cosU1*sinU2-sinU1*cosU2*cosLam))
    
            if sinSigma == 0:
                return 0
    
            cosSigma = sinU1*sinU2+cosU1*cosU2*cosLam
            sigma = math.atan2(sinSigma,cosSigma)
              
            sinAlpha = cosU1*cosU2*sinLam/sinSigma
            cosSqAlpha = 1 - sinAlpha*sinAlpha
            if cosSqAlpha == 0: #catches zero division error if points on equator
                cos2SigmaM = 0
            else:
                cos2SigmaM = cosSigma - 2*sinU1*sinU2/cosSqAlpha
    
            if cos2SigmaM == None:
                cos2SigmaM = 0
    
            C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha))
            lamP = lam
            lam = L+(1-C)*f*sinAlpha*(sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)))
    
            iter_count+=1
    
        uSq = cosSqAlpha*(a*a - b*b)/(b*b)
        A = 1 + uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)))
        B = uSq/1024*(256+uSq*(-128+uSq*(74-47*uSq)))
        deltaSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)-B/6*cos2SigmaM*(-3+4*sinSigma*sinSigma)*(-3+4*cos2SigmaM*cos2SigmaM)))
        s = b*A*(sigma-deltaSigma)
    
        #convert s in meters to miles
        s_miles = s*0.0006213712
        return s_miles
    
    def GetTimeDifference(self, prevDT,currDT):
        """Calculates the difference between to date and time variables. The difference is returned in minutes."""
        timeDelta = currDT - prevDT
        totMinutes = (timeDelta.days*1440) + (timeDelta.seconds/60)
    
        return totMinutes
     
    def AddLines_Segmented(self, dataDicAIS, dataDicAttrs, lOutFieldList, sOutputFC):
        """Creates track polylines and saves them in the ouput feature class.  Track lines are created using a maximum distance and
        maximum time threshold between points."""
        self.printit("\n  Generating track lines...")
        arcpy.SetProgressor("step","Generating track lines...",0,len(dataDicAIS),1)
         
        cur = arcpy.da.InsertCursor(sOutputFC, lOutFieldList)
    
        pnt1 = arcpy.Point()
        pnt2 = arcpy.Point()
        lineArray = arcpy.Array()
    
        for key in sorted(dataDicAIS.keys()):
            iSegCount = 0
            
            
            lAddAttrs = []
            for f in lOutFieldList:
                if f not in ["SHAPE@",sID_Field,"TrackStartTime","TrackEndTime"]:
                    if f in dataDicAttrs[key]:
                        lAddAttrs.append(dataDicAttrs[key][f])
                    else:
                        lAddAttrs.append(None)
            
    
            #list of date time dictionary keys
            dtTimes = sorted(dataDicAIS[key].keys())
    
            #start and end date variables: start date set with the start of any new line.  end date constantly updated with the second point
            #so that whenever the line ends the last datetime will already be set.
            dtSegStart = None
            dtSegEnd = None
    
            #skip MMSI if there is only one point provided
            if len(dtTimes) > 1:
                for i in range(0,len(dtTimes)-1):
                    pnt1.X = dataDicAIS[key][dtTimes[i]][0]
                    pnt1.Y = dataDicAIS[key][dtTimes[i]][1]
    
                    pnt2.X = dataDicAIS[key][dtTimes[i+1]][0]
                    pnt2.Y = dataDicAIS[key][dtTimes[i+1]][1]
    
                    distance = self.GetDistance(pnt1.X,pnt1.Y,pnt2.X,pnt2.Y)
                    timeDiff = self.GetTimeDifference(dtTimes[i],dtTimes[i+1])
    
                    if distance < iMaxDistMiles and timeDiff < iMaxTimeMinutes:
                        if iSegCount == 0:
                            dtSegStart = dtTimes[i]
                            lineArray.add(pnt1)
                            lineArray.add(pnt2)
                        else:
                            lineArray.add(pnt2)
    
                        dtSegEnd = dtTimes[i+1]
                        iSegCount+=1
    
                    #Break the line as the gap exceeds the thresholds
                    else:
                        if lineArray.count > 1:
                            polyline = arcpy.Polyline(lineArray,spatialRef_CURSOR)
                            cur.insertRow((polyline,key,dtSegStart,dtSegEnd) + tuple(lAddAttrs))
                            del polyline
               
                        lineArray.removeAll()
                        dtSegStart = None
                        dtSegEnd = None
    
                        ##reset seg count,since line ended early due to thresholds
                        iSegCount = 0
    
                #end line since end of this MMSI set
                if lineArray.count > 1:
                    polyline = arcpy.Polyline(lineArray,spatialRef_CURSOR)
                    cur.insertRow((polyline,key,dtSegStart,dtSegEnd) + tuple(lAddAttrs))
                    del polyline
    
                lineArray.removeAll()
                dtSegStart = None
                dtSegEnd = None
            arcpy.SetProgressorPosition()
    
        iTotTracks = int(arcpy.GetCount_management(sOutputFC).getOutput(0))
    
        self.printit("    Unique "+sID_Field+"'s= "+self.FormatCounts(len(dataDicAIS)))
        self.printit("    Track lines created = "+self.FormatCounts(iTotTracks))
        arcpy.ResetProgressor()
        del cur
    
    def AddLines_Interval(self, dataDicAIS, dataDicAttrs, lOutFieldList, sOutputFC):
        """Creates track polylines and saves them in the ouput feature class.  Track lines are created based on the specified time interval."""
        self.printit("\n  Generating track lines...")
        arcpy.SetProgressor("step","Generating track lines...",0,len(dataDicAIS),1)
         
        cur = arcpy.da.InsertCursor(sOutputFC, lOutFieldList)
    
        pnt1 = arcpy.Point()
        pnt2 = arcpy.Point()
        lineArray = arcpy.Array()
    
        #create time interval from input parameter hours
        tdInterval = datetime.timedelta(hours=iIntervalHours)
    
        for key in sorted(dataDicAIS.keys()):
            iSegCount = 0
            
            
            lAddAttrs = []
            for f in lOutFieldList:
                if f not in ["SHAPE@",sID_Field,"TrackStartTime","TrackEndTime"]:
                    if f in dataDicAttrs[key]:
                        lAddAttrs.append(dataDicAttrs[key][f])
                    else:
                        lAddAttrs.append(None)
    
            #list of date time dictionary keys
            dtTimes = sorted(dataDicAIS[key].keys())
    
            #start and end date variables: start date set with the start of any new line.  end date constatnly updated with the second point
            #so that whenever the line ends the last datetime will already be set.
            dtSegStart = None
            dtSegEnd = None
    
            #initialize time interval using the first point as the start, and the end based on the selected interval
            #updated as the interval threshold is reached.
            dtIntervalBegin = dtTimes[0]
            dtIntervalEnd = dtIntervalBegin + tdInterval
    
            #skip MMSI if there is only one point provided
            if len(dtTimes) > 1:
                for i in range(0,len(dtTimes)-1):                    
                    pnt1.X = dataDicAIS[key][dtTimes[i]][0]
                    pnt1.Y = dataDicAIS[key][dtTimes[i]][1]
    
                    pnt2.X = dataDicAIS[key][dtTimes[i+1]][0]
                    pnt2.Y = dataDicAIS[key][dtTimes[i+1]][1]
    
                    if dtTimes[i] >= dtIntervalBegin and dtTimes[i+1] <= dtIntervalEnd:
                        if iSegCount == 0:
                            dtSegStart = dtTimes[i]
                            lineArray.add(pnt1)
                            lineArray.add(pnt2)
                        else:
                            lineArray.add(pnt2)
    
                        dtSegEnd = dtTimes[i+1]
                        iSegCount+=1
    
                    #Break the line as the next point is outside the defined interval
                    else:
                        if lineArray.count > 1:                              
                            polyline = arcpy.Polyline(lineArray,spatialRef_CURSOR)
                            cur.insertRow((polyline,key,dtSegStart,dtSegEnd) + tuple(lAddAttrs))
                            del polyline
               
                        lineArray.removeAll()
                        dtSegStart = None
                        dtSegEnd = None
    
                        #update for next time interval
                        dtIntervalBegin = dtIntervalEnd
                        dtIntervalEnd = dtIntervalBegin + tdInterval
    
                        ##reset seg count,since line ended early due to thresholds
                        iSegCount = 0
    
                #end line since end of this MMSI set
                if lineArray.count > 1:
                    polyline = arcpy.Polyline(lineArray,spatialRef_CURSOR)
                    cur.insertRow((polyline,key,dtSegStart,dtSegEnd) + tuple(lAddAttrs))
                    del polyline
                        
                lineArray.removeAll()               
                dtSegStart = None
                dtSegEnd = None
                dtIntervalBegin = None
                dtIntervalEnd = None
            arcpy.SetProgressorPosition()
    
        iTotTracks = int(arcpy.GetCount_management(sOutputFC).getOutput(0))
    
        self.printit("    Unique "+sID_Field+"'s= "+self.FormatCounts(len(dataDicAIS)))
        self.printit("    Track lines created = "+self.FormatCounts(iTotTracks))
        arcpy.ResetProgressor()
        del cur

    
    
    def CreateTracks(self, parameters):  
        """Main function that calls all of the other functions.  Determines which functions are run, based on the user selected parameters."""
                
        #########################################################################################################
        ## Input Paramaters
        global sInputFile, sID_Field, sDT_Field, sOutWorkspace, sOutName
        global sLineMethod, iMaxTimeMinutes, iMaxDistMiles, iIntervalHours
        global lFields, spatialRef_CURSOR
        
        sInputFile = parameters[0].value.value
        sID_Field = parameters[1].value.value
        sDT_Field = parameters[2].value.value
        sOutWorkspace = parameters[3].value.value
        sOutName = parameters[4].value
        sLineMethod = parameters[5].value
        iMaxTimeMinutes = int(parameters[6].value)
        iMaxDistMiles = int(parameters[7].value)
        iIntervalHours = int(parameters[8].value)
        
        vt = parameters[9].value
        lFields = []
        if vt != None:      
            for i in range(vt.rowcount):
                fields = arcpy.ListFields(sInputFile)
                for f in fields:
                    if f.name == vt.getvalue(i, 0):
                        lFields.append(f)
        
        prjFile = os.path.join(arcpy.GetInstallInfo()["InstallDir"],"Coordinate Systems/Geographic Coordinate Systems/World/WGS 1984.prj")
        spatialRef_WGS84 = arcpy.SpatialReference(prjFile)
        spatialRef_input = arcpy.Describe(sInputFile).spatialReference
        
        #global spatialRef_CURSOR
        if spatialRef_WGS84.name != spatialRef_input.name:
            spatialRef_CURSOR = spatialRef_WGS84
        else:
            spatialRef_CURSOR = None
        #########################################################################################################
        

        #Create Dictionary of broadcast data Key = ID, Value = Dictionary of {DateTime:[x,y,voyageID]}
        dataDicAIS, dataDicAttrs = self.CreateDataDict()
    
        sOutputFC = sOutWorkspace+"\\"+sOutName
         
        #Create output feature class          
        lOutFieldList = self.CreateOuputDataset()
              
        #Add Lines to FC
        if sLineMethod == "Maximum Time and Distance":
            self.AddLines_Segmented(dataDicAIS, dataDicAttrs, lOutFieldList, sOutputFC)
        elif sLineMethod == "Time Interval":
            self.AddLines_Interval(dataDicAIS, dataDicAttrs, lOutFieldList, sOutputFC)       
              
        self.printit("\n\n")
        


class AIS_CSVtoFGDB(object):
    def __init__(self):
        self.label = "AIS CSV to File Geodatabase"
        self.description = "Takes an ASCII CSV file of AIS data and converts it to a point feature class within a file geodatabase"
        self.canRunInBackground = False

    def getParameterInfo(self):
        param0 = arcpy.Parameter(
            displayName="Input CSV File",
            name="Input_CSV",
            datatype="DEFile",
            parameterType="Required",
            direction="Input")
        
        param0.filter.list = ['csv']        

        param1 = arcpy.Parameter(
            displayName="Output Point FeatureClass",
            name="Output_FeatureClass",
            datatype="DEFeatureClass",
            parameterType="Required",
            direction="Output")
                
        params = [param0, param1]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        if parameters[0].value:
            if not parameters[1].altered:
                sDefaultName = os.path.splitext(os.path.basename(parameters[0].value.value))[0]
                if arcpy.env.workspace == None:
                    if arcpy.env.scratchWorkspace == None:
                        sWkspc = arcpy.env.scratchGDB
                    else:
                        sWkspc = arcpy.env.scratchWorkspace
                else:
                    sWkspc = arcpy.env.scratchWorkspace
                parameters[1].value = os.path.join(sWkspc,sDefaultName)
        return
    
    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
                
        if parameters[1].value:
            wkspc = os.path.dirname(parameters[1].value.value)
            desc = arcpy.Describe(wkspc)
            if desc.workspaceType != "LocalDatabase":
                parameters[1].setErrorMessage("Output Point FeatureClass can only be created in a FileGeodatabase")
                
        return

    def execute(self, parameters, messages):
        self.CreateOuputDataset(parameters)
        self.WritePoints(parameters)
        
        return
    
    
    def CreateOuputDataset(self, parameters):
        """Creates the empty output point feature class, and sets up all the AIS attribute fields."""
        arcpy.env.overwriteOutput = True
        sr = arcpy.SpatialReference(os.path.join(arcpy.GetInstallInfo()["InstallDir"],
                                                 r"Coordinate Systems/Geographic Coordinate Systems/North America/NAD 1983.prj"))
        
        #create a feature class in memory, to use as a template
        arcpy.CreateFeatureclass_management('in_memory','template',"POINT","","DISABLED","DISABLED",sr)
        tmp = os.path.join('in_memory', 'template')
    
        #add all the required fields
        arcpy.AddField_management(tmp,"MMSI","LONG")
        arcpy.AddField_management(tmp,"BaseDateTime","DATE")
        arcpy.AddField_management(tmp,"SOG","DOUBLE")
        arcpy.AddField_management(tmp,"COG","DOUBLE")
        arcpy.AddField_management(tmp,"Heading","DOUBLE")
        arcpy.AddField_management(tmp,"VesselName","TEXT","","",32)
        arcpy.AddField_management(tmp,"IMO","TEXT","","",15)
        arcpy.AddField_management(tmp,"CallSign","TEXT","","",8)
        arcpy.AddField_management(tmp,"VesselType","LONG")
        arcpy.AddField_management(tmp,"Status","TEXT","","",64)
        arcpy.AddField_management(tmp,"Length","DOUBLE")
        arcpy.AddField_management(tmp,"Width","DOUBLE")
        arcpy.AddField_management(tmp,"Draft","DOUBLE")
        arcpy.AddField_management(tmp,"Cargo","LONG")
    
        #create the real output feature class based on the template
        arcpy.CreateFeatureclass_management(os.path.dirname(parameters[1].value.value),os.path.basename(parameters[1].value.value),
                                            "POINT",tmp,"DISABLED","DISABLED",sr)
        
        #delete the temporary template dataset in memory
        arcpy.Delete_management(tmp)
        del tmp
        
        arcpy.AddMessage("Output point feature class created.")
        
        
    def WritePoints(self, parameters):
        """Reads the input CSV file and writes all the points and attribute data to the output point feature class."""
        
        fields = ['SHAPE@','MMSI','BaseDateTime','SOG','COG','Heading','VesselName','IMO','CallSign',
                  'VesselType','Status','Length','Width','Draft','Cargo']
        
        with arcpy.da.InsertCursor(parameters[1].value, fields) as cur: 
            #first gets the total number of points, in the CSV to show progress
            iTotal = 0
            with open(parameters[0].value.value) as f:
                for l in f:
                    iTotal += 1
                    
            p = int(math.log10(iTotal))
            if not p:
                p = 1
            increment = int(math.pow(10, p - 1))
            
            arcpy.SetProgressor("step", "Writing points to output FeatureClass: {:,} of {:,} complete".format(0,iTotal),0, iTotal-1, increment)
            
            #read each line and format/prepare each attribute    
            with open(parameters[0].value.value) as f:
                i = 0
                for l in f:
                    if i == 0:
                        i+=1
                        continue
                    
                    parts = (l.strip()).split(",")                    
                    
                    pnt = arcpy.Point(float(parts[3]), float(parts[2]))
                    mmsi = self.ToInt(parts[0])
                    dtDT = datetime.datetime.strptime(parts[1],"%Y-%m-%dT%H:%M:%S") #Converting the UTC date string to the ESRI standard format
                    sog = self.ToFloat(parts[4])
                    cog = self.ToFloat(parts[5])
                    heading = self.ToFloat(parts[6])
                    vname = self.ToString(parts[7])
                    imo = self.ToString(parts[8])
                    callsign = self.ToString(parts[9])
                    vtype = self.ToInt(parts[10])
                    status = self.ToString(parts[11])
                    length = self.ToFloat(parts[12])
                    width = self.ToFloat(parts[13])
                    draft = self.ToFloat(parts[14])
                    cargo = self.ToInt(parts[15])
                    
                    new_row = [pnt,mmsi,dtDT,sog,cog,heading,vname,imo,callsign,vtype,status,length,width,draft,cargo]
                    cur.insertRow(new_row)
                    
                    i+=1
                    
                    if (i % increment) == 0:
                        arcpy.SetProgressorPosition(i)
                        arcpy.SetProgressorLabel("Writing points to output FeatureClass: {:,} of {:,} complete".format(i,iTotal))
                        
                arcpy.SetProgressorPosition(i)

        arcpy.ResetProgressor()       
        arcpy.AddMessage("Process Complete!")
        arcpy.AddMessage("Total AIS Points Created = " + str(iTotal))
        
    def ToInt(self, sInput):
        """Simple function to convert the input CSV string data to integer, or NULL if the string is empty."""
        if sInput == "":
            return None
        else:
            return int(sInput)
        
    def ToFloat(self, sInput):
        """Simple function to convert the input CSV string data to float, or NULL if the string is empty."""
        if sInput == "":
            return None
        else:
            return float(sInput)        
            
    def ToString(self, sInput):
        """Simple function to replace empty input CSV strings with NULL."""
        if sInput == "":
            return None
        else:
            return sInput        